# 2008.02.24 23:21 Logs are now lower level objects than Journals
# Journals are 2 dimensional objects,
#   - 1 dimension being time
#   - the other being tags
#
# 2008.11.04 16:42
# seems more like 3 dimensional:  time, tags, data

import re, codecs
from datetime import datetime

#import common

#from tag import Tag
from association import Association
#from timestamp import time_format
from log import Log
from entry import Entry
from timestamp import Timestamp

def phraseUnicode2ASCII(message):
    """
    Works around the built-in function str(message) which aborts when non-ASCII
    unicode characters are given to it.

    Modified from http://mail.python.org/pipermail/python-list/2002-June/150077.html
    """
    try:
        newMsg = message.encode('ascii')
    except (UnicodeDecodeError, UnicodeEncodeError):
       chars=[]
       for uc in message:
          try:
             char = uc.encode('ascii')
             chars.append(char)
          except (UnicodeDecodeError, UnicodeEncodeError):
             pass
       newMsg = ''.join(chars)
    return newMsg.strip()

class Journal:

    def __init__(self, name=None):
        #can be filename:
        #*2009.03.09 19:18:57  is this used anywhere?
        self.name = name

        #entries are variable increments of time, associated with objects
        #keys must be start timestamp
        #
        #this dict stores actual Entry objects for journal
        self.entries = {}

        #both of the following could be generalized as Associations
        #maybe rework Tag object to be that

        #tags are objects
        #keys must be tag name
        #
        #shouldn't it just reference a list of entries w/ those tags?
        #that is essentially what tag is,
        #but tag only keeps a list of entry times associated with tag
        #not actual entries
        #this makes it easier to see if an entry has been associated already
        self.tags = Association()

        #different entries can have the same data
        #would be useful to have an index for that
        #self.datas = {}
        self.datas = Association()

        #as entries are added, can add a timestamp here
        #then, as desired, can be re-exported in order
        #*2009.03.23 18:53:07
        #journal should be a subclass of list
        #then order could be stored there
        self.order = []
        
    def to_file(self, filename=None):
        """
        >>> from entry import Entry
        >>> e = Entry("test entry")
        >>> j.add_entry(e)
        >>> j.to_file("sample_log2.txt")
        >>> k = Journal()
        >>> k.from_file("sample_log2.txt")
        >>> len(k.entries)
        2
        """
        if filename:
            l = Log(filename)
        elif self.name:
            l = Log(self.name)
        else:
            print "No name to save Journal to"
            exit()

        #l.from_journal(self, holder, entry)
        l.from_entries(self.to_entries())
        l.to_file()
        l.close()

    def to_entries(self, sort='original', all_placeholders=True):
        """
        return a list of entries the make up the current journal

        can specify order='original'
        to keep the original order that the entries were added to the journal

        'chronological'

        if not all entries are wanted, see self.limit()
        """
        entry_times = self.entries.keys()

        #just want to make sure things are consistent in the journal object
        assert len(entry_times) == len(self.order)

        if sort == "original":
            #print "ORIGINAL"
            #self.order.reverse()
            entry_times = self.order
        else:
            #print "CHRONOLOGICAL"
            entry_times.sort()
            entry_times.reverse()

        entries = []
        #print "Placeholders: %s" % placeholders_only
        #print "Entries: %s" % entries_only

        for et in entry_times:
            entry = self.entries[et]
            #if the entry is a placeholder, and we want placeholders, add it
            if entry.placeholder and all_placeholders:
                entries.append(entry)

            #otherwise it must be a normal entry.  
            else:
                entries.append(entry)
        return entries

    def flatten(self, filename=None):
        """
        save without timestamps (or tags?)
        """
        if filename:
            f = codecs.open(filename, 'w', encoding='utf-8')
        elif self.name:
            f = codecs.open(self.name, 'w', encoding='utf-8')
        else:
            print "No name to save file to"
            exit()

        entries = self.to_entries()
        flat = ''
        for e in entries:
            flat += e.render_data()

        f.write(flat)
        f.close()


    #formerly: add_log_to_journal, add_file
    def from_file(self, log_name, add_tags=[]):
        """
        adds a log file to the journal object currently in memory

        >>> from journal import *
        >>> j = Journal()
        >>> j.from_file("sample_log.txt")
        >>> len(j.entries)
        1
        """

        l = Log(log_name)
        l.from_file(log_name)
        
        #if our name wasn't originally initialized, go ahead and set it:
        if not self.name:
            self.name = log_name

        entries = l.to_entries(add_tags)
        self.from_entries(entries)
        
        l.close()

    def from_entries(self, entries):
        """
        loop over a list of entries to add/update each one to the journal
        """
        for e in entries:
            self.update_entry(e)
        
    def _add_entry(self, entry, position=None):
        """
        assumes that there is nothing in the journal at the entry time
        this check happens in update_entry
        as of 2008.09.23 13:31 _add_entry is a private method of Journal

        ADD ENTRY WILL OVERWRITE EXISTING ENTRIES!!

        this is the base case for adding an entry
        adds the entry object to the journal's list of entries
        """
        #current_time = entry.created_at.strftime(time_format)
        current_time = str(Timestamp(time=entry.created_at))
        self.entries[current_time] = entry

        #then add it to the Journal's Tags
        #*2009.03.15 09:02:27
        #is this called anywhere else?
        #I couldn't see anywhere...
        #therefore doesn't really need to be a separate function
        #self.update_tags(entry)

        #add the entry's tags to the journal's list of tags
        current_tags = entry.tags
        for t in current_tags:
            self.tags.associate_entry(entry, t)

        #add the entry's data to the journal's data index / association

        #check for special case of entry data being a list of items
        # (used for many to many relationships)
        #this doesn't get any whitespace, which could be included
        #data_list = re.compile("^\[\w*\]$")
        #match anything that is not a [ or a ]:
        data_list = re.compile("^\[[^\]\[]*\]$")
        if data_list.match(entry.data):
            items = entry.data[1:-1].split()
            #split the data into it's elements
            for i in items:
                if i:
                    self.datas.associate_entry(entry, i)
        else:
            self.datas.associate_entry(entry, entry.data)

        #to keep track of the original order
        #that the entries were added to the journal
        #this should usually correspond with the order in a log
        #if journal consists of entries from one log
        if not current_time in self.order:
            if position != None:
                self.order.insert(position, current_time)
            else:
                self.order.append(current_time)


    def replace_entry(self, entry):
        """
        accepts a new entry
        and uses it to find and then remove the original one
        add the new one to the journal
        thereby replacing the original
        """
        current_time = str(Timestamp(time=entry.created_at))
        existing = self.entries[current_time]
        self.remove_entry(existing)
        self._add_entry(entry)

        
    def update_entry(self, entry, position=None, keep=None):
        """
        checks if an entry already exists in the journal
        if exists, sees if that entry is a placeholder
        replaces it if placeholder

        OTHERWISE:
        uses 'keep' argument and checks the existing one to determine what to do
        this would be the place to attempt merging individual moments
        """
        current_time = str(Timestamp(time=entry.created_at))
        #current_time = entry.created_at.strftime(time_format)

        #print "Adding Entry: %s" % current_time
        if not self.entries.has_key(current_time):
            self._add_entry(entry, position)
        else:
            #it must have *something* in that time slot,
            existing = self.entries[current_time]
            #need to see if we have a placeholder entry in the current key's pos
            if existing.is_placeholder():
                #go ahead and replace it, that's why we're adding:
                #print "REPLACING PLACEHOLDER: %s" % (current_time)
                self.replace_entry(entry)
            elif ( (existing.data == entry.data) and
                   (existing.tags == entry.tags) ):
                print "DUPE, but tags and data are same... skipping"
            elif existing.data == entry.data:
                #tags must differ... those are easy to merge:
                print "only TAGS differ"

            elif keep and keep == '>':
                #if we want to keep '<', that is the one already here. (skip)
                #otherwise:
                self.remove_entries( [existing] )
                self._add_entry(entry)
            elif keep:
                print "keeping '<', the first version added"
                
            else:
                print "DUPLICATE ENTRY EXISTS AT: %s" % (current_time)
                print "CONFLICT FOUND!!!! MANUAL MERGE REQUIRED!"
                print "< : " + str(self.entries[current_time].source_file)
                print "< : " + str(self.entries[current_time].tags)
                for line in self.entries[current_time].data.split('\n'):
                    print "< : " + phraseUnicode2ASCII(line)
                print ""

                print "> : " + str(entry.source_file)
                print "> : " + str(entry.tags)
                for line in entry.data.split('\n'):
                    print "> : " + phraseUnicode2ASCII(line)
                print ""

    def remove_entry(self, entry, destination_log=None):
        """
        for each entry
          scan through all tags to see if they reference that entry,
          if they do, remove the reference
          then remove the entry from the journal.

        """
        text_time = str(Timestamp(time=entry.created_at))
        #text_time = e.created_at.strftime(time_format)

        self.tags.remove_entry(entry)
        self.datas.remove_entry(entry)

        #this does not delete actual e, just its key in entries dict
        #need to do this before adding a placeholder in.
        #add will overwrite
        #(could actually leave del until later in that case)
        del self.entries[text_time]

        #make_placeholder=False
        #destination_log is used to see if we need to make a placeholder
        if destination_log:
            ph = entry.to_placeholder(destination_log)
            #would be better to re-add the entry the right way,
            #this way journal tags will get associated too
            self.add_entry(ph)

        #could consider deleting actual e object, but it's no longer
        #referenced by the journal either way at this point

    def remove_entries(self, entries, destination_log=None):
        """
        take a list of entry objects,
        remove each one
        """
        for e in entries:
            self.remove_entry(e, destination_log)

    def extract(self, tag_list, etype="intersect"):
        """
        saving journal (with missing entries) left to caller (if desired)
        """
        
        entries = []
        if etype == "union":
            entries = self.union_tags(tag_list)
        elif etype == "intersect":
            entries = self.intersect_tags(tag_list)
        else:
            print "Unknown type of extraction!!!"

        #sending fname creates place holder entries in the journal
        #not sure if we still need to do that [2008.11.04 16:58]
        #(dir, fname) = os.path.split(ofile)
        #self.remove_entries(entries, fname)
        self.remove_entries(entries)

        #rather than save here, lets just return the list of entries
        #then they can be added to another journal instance
        #or written directly to log by caller as they were here.
        return entries


    def timelists_from_tags(self, tag_names):
        """
        aka make_entry_time_lists_from_tags
        take a list of tag names,
        return a list of actual tag objects from this journal.
            [could possibly have the same tag in different journals in memory]
        """
        if tag_names:
            list_of_timelists = []
            for name in tag_names:
                if self.tags.has_key(name):
                    list_of_timelists.append(self.tags[name])
            return list_of_timelists
        else:
            return [self.entries.keys()]

    def intersect_tags(self, tag_name_list):
        """
        similar to union_tags,
        but only return those entries with all specified tags

        only return entries that show up in all lists
        """
        list_of_lists = self.timelists_from_tags(tag_name_list)

        #merge the tag's entries into one array (avoid dupes)
        entry_list = []
        first_tag = list_of_lists.pop()
        time_set = set(first_tag)
        for l in list_of_lists:
            time_set = time_set.intersect(set(l))

        time_list = list(time_set)
        time_list.sort()
        time_list.reverse()

        for time in time_list:
            entry = self.entries[time]
            entry_list.append(entry)

        return entry_list

    def union_tags(self, tag_name_list):
        """
        takes a list of tags
        union all tag entries
        return a list of actual entry objects
        """
        list_of_lists = self.timelists_from_tags(tag_name_list)

        #merge the tag's entries into one array (avoid dupes)
        entry_list = []
        time_set = set()
        for t in list_of_lists:
            time_set = time_set.union(set(t))

        time_list = list(time_set)
        time_list.sort()
        time_list.reverse()

        for time in time_list:
            entry = self.entries[time]
            entry_list.append(entry)

        return entry_list

    def show_entries(self):
        entries = self.to_entries()
        for e in entries:
            print e.render()
    
    def make_graph(self):
        g = graph.Graph(self.name, self.j)
        g.make_links()
        g.write_graph2()

    def limit(self, start, end=datetime.now()):
        times = self.entries.keys()
        times.sort()

        matches = []
        for t in times:
            pytime = Timestamp(t).time
            if (pytime >= start) and (pytime <= end):
                matches.append(self.entries[t])
        return matches

if __name__ == "__main__":

    import doctest

    doctest.testmod()
